# tclgeomap_procs.tcl --
#
#	This file provides the tclgeomap_procs package, which offers procedures
#	for interacting with geographic data in Tcl.
#  
#  Copyright (c) 2004 Gordon D. Carrie.  All rights reserved.
#  
#  Licensed under the Open Software License version 2.1
#  
#  Please address questions and feedback to user0@tkgeomap.org
# 
# @(#) $Id: tclgeomap_procs.tcl,v 1.26 2007/06/14 21:05:18 tkgeomap Exp $

package require tclgeomap 2
package provide tclgeomap_procs 2.11.4

namespace eval geomap {
    namespace export georadius latlonok mapptok lonbtwn gclcross rotatpt \
	    scalept domnlat domnlon domnlonpt gwchlon gwchpt dmstodec dectodm \
	    dectodms cartg projection lnarr place
}

namespace eval geomap {
    namespace export latitude longitude circle fillsegment
}

# geomap::latitude
#
#     Obtain the latitude from a point.
#
# Arguments:
#     latLon	a geographic location given as {latitude longitude}
# Results:
#     The latitude of the point.

proc geomap::latitude {latLon} {
    if ![latlonok $latLon] {
	error "$latLon not a {latitude longitude}"
    }
    return [lindex $latLon 0]
}

# geomap::longitude
#
#     Obtain the longitude from a point.
#
# Arguments:
#     latLon	a geographic location given as {latitude longitude}
# Results:
#     The longitude of the point.

proc geomap::longitude {latLon} {
    if ![latlonok $latLon] {
	error "$latLon not a {latitude longitude}"
    }
    return [lindex $latLon 1]
}

# geomap::latlon --
#
#	Obtain a {latitude longitude} value, converting from degree minute
#	second representation if necessary.
#
# Arguments:
#	latLon	- a geographic location, specified as decimal
#		  {latitude longitude} or {deg min sec NorS deg min sec EorW}.
#
# Results:
#	If latLon argument is a decimal {latitude longitude}, return value is
#	latLon.  If latLon argument is {deg min sec NorS deg min sec EorW},
#	return value is corresponding decimal {latitude longitude}.  Otherwise,
#	the procedure generates an error.

proc geomap::latlon {latLon} {
    if [latlonok $latLon] {
	return $latLon
    } else {
	return [dmstodec $latLon]
    }
}

# geomap::prt_time --
#
#	Extract elements from a time.
#
# Arguments:
#	time	- time of form {year month day hour minute second}
#	year	- name of variable to receive year value from time.
#	month	- name of variable to receive month value from time.
#	day	- name of variable to receive day value from time.
#	hour	- name of variable to receive hour value from time.
#	minute	- name of variable to receive minute value from time.
#	second	- name of variable to receive second value from time.

proc geomap::prt_time {time year month day hour minute second} {
    upvar $year y
    upvar $month mo
    upvar $day d
    upvar $hour h
    upvar $minute mi
    upvar $second s
    set y [expr {int([lindex $time 0])}]
    set mo [expr {int([lindex $time 1])}]
    set d [expr {int([lindex $time 2])}]
    set h [expr {int([lindex $time 3])}]
    set mi [expr {int([lindex $time 4])}]
    set s [expr {double([lindex $time 5])}]
}

# geomap::circle
#
#     Create list of points at distance radius from center
#
# Arguments:
#     center	geographic point at the center of the circle
#     radius	radius of the circle.  Default is gsdeg (great circle degrees)
#     unit	(optional) unit for radius
# Results:
#     A list of geographic points at distance radius from center

proc geomap::circle {center radius {unit gsdeg}} {
    for {set az 360} {$az >= 0} {incr az -3} {
	lappend circle [place step $center $az $radius $unit]
    }
    return $circle
}

# geomap::fillsegment
#
#	This procedure fills in a segment with points.  This helps it maintain
#	a great circle shape in various projections.
#
# Arguments:
#	p1	endpoint
#	p2	endpoint
#	d	distance between points added between p1 and p2
# Results:
#	A list of geographic points going from p1 to p2

proc geomap::fillsegment {p1 p2 {d 2.0}} {
    if {$d <= 0} {
        error "Distance between points must be greater than 0.0"
    }
    set azRng [place azrng $p1 $p2]
    set az [lindex $azRng 0]
    set rng [lindex $azRng 1]
    if {$rng <= $d} {
        return [list $p1 $p2]
    }
    lappend ln $p1
    for {set r $d} {$r < $rng} {set r [expr {$r + $d}]} {
        set p [place step $p1 $az $r]
        lappend ln $p
    }
    if {[longitude $p] != [longitude $p2] || [latitude $p] != [latitude $p2]} {
        lappend ln $p2
    }
    return $ln
}

# geomap::grid_list --
#
#	This procedure returns a list of grid lines.
#
# Arguments:
#	See the tclgeomap_procs (n) man page.
# Results:
#	See the tclgeomap_procs (n) man page.

proc geomap::grid_list {{s 15.0}} {
    if {$s <= 0.0} {
        error "Grid spacing must be positive"
    }
    set dl 2.0

    # Make latitude circles

    for {set lat [expr {- $s * int(90.0 / $s)}]} {$lat < 90.0} \
	    {set lat [expr {$lat + $s}]} {
	for {set lon0 [expr {- $s * int(180.0 / $s)}]} {$lon0 < 180.0} \
		{set lon0 [expr {$lon0 + $s}]} {
	    set lon1 [expr {$lon0 + $s}]
	    for {set lon $lon0} {$lon <= $lon1} {set lon [expr {$lon + $dl}]} {
		lappend grid [list $lat $lon]
	    }
	    if {$lon != $lon1} {
		lappend grid [list $lat $lon1]
	    }
	    lappend grids $grid
	    unset grid
	}
    }

    # Make meridians

    for {set lon [expr {- $s * int(180.0 / $s)}]} {$lon < 180.0} \
	    {set lon [expr {$lon + $s}]} {
	for {set lat0 [expr {- $s * int(90.0 / $s)}]} {$lat0 < 90.0} \
		{set lat0 [expr {$lat0 + $s}]} {
	    set lat1 [expr {$lat0 + $s}]
	    for {set lat $lat0} {$lat <= $lat1} {set lat [expr {$lat + $dl}]} {
		lappend grid [list $lat $lon]
	    }
	    if {$lat != $lat1} {
		lappend grid [list $lat1 $lon]
	    }
	    lappend grids $grid
	    unset grid
	}
    }

    return $grids
}

# geomap::ocean_list --
#
#	This procedure returns a list of small closed squares that cover the
#	earth.  
#
# Arguments:
#	See the tclgeomap_procs (n) man page.
# Results:
#	See the tclgeomap_procs (n) man page.

proc geomap::ocean_list {{s 15.0}} {
    if {$s <= 0.0} {
        error "Grid spacing must be positive"
    }
    set dl 2.0
    for {set lat0 [expr {- $s * int(90.0 / $s)}]} {$lat0 < 90.0} \
	    {set lat0 [expr {$lat0 + $s}]} {
	for {set lon0 [expr {- $s * int(180.0 / $s)}]} {$lon0 < 180.0} \
		{set lon0 [expr {$lon0 + $s}]} {
	    set lat1 [expr {$lat0 + $s}]
	    set lon1 [expr {$lon0 + $s}]
	    for {set lat $lat0;set lon $lon0} {$lon <= $lon1} \
		    {set lon [expr {$lon + $dl}]} {
		lappend ocean [list $lat $lon]
	    }
	    for {set lat $lat0;set lon $lon1} {$lat <= $lat1} \
		    {set lat [expr {$lat + $dl}]} {
		lappend ocean [list $lat $lon]
	    }
	    for {set lat $lat1;set lon $lon1} {$lon >= $lon0} \
		    {set lon [expr {$lon - $dl}]} {
		lappend ocean [list $lat $lon]
	    }
	    for {set lat $lat1;set lon $lon0} {$lat >= $lat0} \
		    {set lat [expr {$lat - $dl}]} {
		lappend ocean [list $lat $lon]
	    }
	    lappend ocean [list $lat0 $lon0]
	    lappend oceans $ocean
	    unset ocean
	}
    }
    return $oceans
}

namespace eval geomap {

    # This is a list of recognized projections.

    variable ProjNames
    set ProjNames {CylEqDist Mercator CylEqArea LambertConfConic \
	    LambertEqArea Stereographic PolarStereographic Orthographic}

    # These variables identify type of reference for each projection.

    variable ProjRef
    set ProjRef(CylEqDist) point
    set ProjRef(Mercator) longitude
    set ProjRef(CylEqArea) longitude
    set ProjRef(LambertConfConic) point
    set ProjRef(LambertEqArea) point
    set ProjRef(Stereographic) point
    set ProjRef(PolarStereographic) point
    set ProjRef(Orthographic) point

    # These variables identify domain for each projection.

    variable ProjDomain
    set ProjDomain(CylEqDist) globe
    set ProjDomain(Mercator) globe
    set ProjDomain(CylEqArea) globe
    set ProjDomain(LambertConfConic) globe
    set ProjDomain(LambertEqArea) hemisphere
    set ProjDomain(Stereographic) hemisphere
    set ProjDomain(PolarStereographic) hemisphere
    set ProjDomain(Orthographic) hemisphere
}

# geomap::projections --
#
#	This query procedure returns a list of known projections.
#
# Arguments:
#	None.
#
# Results:
#	Return value is a list of projections.

proc geomap::projections {} {
    return [list CylEqDist Mercator CylEqArea LambertConfConic LambertEqArea \
	    Stereographic PolarStereographic Orthographic]
}

# geomap::proj_info --
#
#	This query procedure returns projection properties.
#
# Arguments:
#	proj_name	- projection name, which should be in the return
#			  value from projections.
#
# Results:
#	Return value is a list of form {reference domain}, where reference
#	is the type of reference the projection uses, either "point" or
#	"longitude", and domain is how much of the Earth the projection
#	transforms, either "globe" or "hemisphere".

proc geomap::proj_info {proj_name} {
    variable ProjRef
    variable ProjDomain

    return [list $ProjRef($proj_name) $ProjDomain($proj_name)]
}
